/*
	APL Compiler

        sam sirlin

	Run time system
	routines having to do with matrices
        calls linpack routines for now...

        take, drop added


 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 02111-1307, USA.

*/

#include <stdio.h>
#include <math.h>
#include "aplc.h"
#include "memalloc.h"
/*#include "f2c.h"*/

#ifdef USE_SYSTEM_LAPACK
extern void zgelsx_( int *M, int *N, int *NRHS, double *A, int *LDA, 
                     double *B, int *LDB, int *JPVT, 
                     double *RCOND, int *RANK,
                     double *WORK, double *RWORK, int *INFO );
#endif

/* set to 0 or 1 */
#define RMDEBUG 0

/* local function declarations */
static int imax(int i, int j);
#if RMDEBUG
static void prtmat(union mp_struct a, int m, int n);
#endif

/* maximum of 2 integer arguments - define as function so it doesn't
   evaluate its arguments multiple times */
static int
imax(int i, int j)
{
  return ((i > j) ? i : j);
}

/* could be for long int */
static la_int
imaxl(la_int i, la_int j)
{
  return ((i > j) ? i : j);
}

#if RMDEBUG
/* 
   printout matrix in mp_struct, for debugging 
   print out a as an mxn matrix
*/
static void
prtmat(union mp_struct a, int m, int n)
{
  union mp_struct temp;
  int i, j;

  temp = a;
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      /* printf("%f ", *temp.rp++);*/
      printf("%10.5g ", *temp.rp++);
    }
    printf("\n");
  }
  return;
}

static void
prtmatl(union mp_struct a, la_int m, la_int n)
{
  union mp_struct temp;
  int i, j;

  temp = a;
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      /* printf("%f ", *temp.rp++);*/
      printf("%10.5g ", *temp.rp++);
    }
    printf("\n");
  }
  return;
}

static void
prtmatd(double *a, la_int m, la_int n)
{
  int i, j;

  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      /* printf("%f ", *temp.rp++);*/
      printf("%10.5g ", *a++);
    }
    printf("\n");
  }
  return;
}

static void
prtmatz(double *a, la_int m, la_int n)
{
  int i, j, k;

  k = 0;
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      /* printf("%f ", *temp.rp++);*/
      printf("(%10.5g,%10.5g) ", a[k], a[k+1]);
      k+=2;
    }
    printf("\n");
  }
  return;
}
#endif

/* change from aplc to fortran form for real matrix 
   - need space allocated already
*/
static void
matrix_f_aplc(double *ap, struct trs_struct *c)
{
  int i,j, n;

  if (c->rank < 2)
    n = 1;
  else
    n = c->shape[1];
  switch (c->type) {
  default:
    aplc_error("domain aplc_error []");
    return;
    break;
  case APLC_BOOL:
  case APLC_INT:
    for (i = 0, j = 0; i < c->size; i++, j += n) {
      if (j >= c->size)
        j %= c->size - 1;
      *ap++ = c->value.ip[j];
    }
    break;
  case APLC_REAL:
    for (i = 0, j = 0; i < c->size; i++, j += n) {
      if (j >= c->size)
        j %= c->size - 1;
      *ap++ = c->value.rp[j];
    }
    break;
  }
  return;
}

/* change from aplc to fortran form for complex matrix 
   - need space allocated already
*/
static void
matrix_z_f_aplc(double *ap, struct trs_struct *c)
{
  int i,j, n;

  if (c->rank < 2)
    n = 1;
  else
    n = c->shape[1];
  /* expect complex c, but take int or real */
  switch (c->type) {
  default:
    aplc_error("domain aplc_error []");
    return;
    break;
  case APLC_BOOL:
  case APLC_INT:
    for (i = 0, j = 0; i < c->size; i++, j += n) {
      if (j >= c->size)
        j %= c->size - 1;
      *ap++ = c->value.ip[j];
      *ap++ = 0;
    }
    break;
  case APLC_REAL:
    for (i = 0, j = 0; i < c->size; i++, j += n) {
      if (j >= c->size)
        j %= c->size - 1;
      *ap++ = c->value.rp[j];
      *ap++ = 0;
    }
    break;
  case APLC_COMPLEX:
    for (i = 0, j = 0; i < c->size; i++, j += n) {
      if (j >= c->size)
        j %= c->size - 1;
      *ap++ = c->value.rp[2*j];
      *ap++ = c->value.rp[2*j+1];
    }
    break;
  }
  return;
}


/* real matrix inverse 
   - requires ap in fortran form 
   - result returned in aplc form */
static void
matinv_real(struct trs_struct *res, double *ap, la_int m, la_int n)
{
  double rcond;
  double *work;
  union mp_struct b;
  union mp_struct c;
  union mp_struct mptemp;
  int size, lsize;
  int i, j;
  la_int mn, nw;
  la_int *jpvt; 
  la_int nrhs, lda, ldb, rank, info;

  lda = m;
  ldb = (m < n) ? n : m;
  nrhs = m;
  size = m * n;
  rcond = DBL_EPSILON;

  /* setup workspace */
  mn = (m < n) ? m : n;
  nw = imaxl(mn + 3 * n, 2 * mn + nrhs);
  aplc_vectalloc(&mptemp, (int) nw, APLC_REAL);
  work = mptemp.rp;

  jpvt = (la_int *) APLC_MALLOC( n*sizeof(la_int) );
  for (i = 0; i < n; ++i) {
    jpvt[i] = 0;
  }
  /* printf("\n mn %d, nw %d\n", mn, nw); */

  /* put identity in b  */
  lsize = ldb * nrhs;
  aplc_vectalloc(&b, lsize, APLC_REAL);
  c = b;
  j = nrhs + 1;
  for (i = 0; i < lsize; i++) {
    if (0 == i % j)
      *c.rp++ = 1.0;
    else
      *c.rp++ = 0.0;
  }

#if RMDEBUG
  /* sws  debugging printout */
  printf("\naplc_domino debugging\n");
  printf("\n a\n");
  prtmatd(ap,n,m);
  printf("\n b\n");
  prtmatl(b,nrhs,ldb);
  printf("\n m %ld, n %ld, nrhs %ld, lda %ld, ldb %ld, rcond %g\n",
	 m,n,nrhs,lda,ldb, rcond);
#endif

  /* solve */
  dgelsx_(&m, &n, &nrhs, ap, &lda, b.rp, &ldb, jpvt, &rcond,
	  &rank, work, &info);
  if (rank != n) {
    /* don't force exit for this error, just notify the user */
    fprintf(aplcerr, "\ndgelsx info = %ld, rank= %ld\n", (long)info, (long)rank);
    aplc_error("domain aplc_error [domino - matrix not of full right rank]");
  }
#if RMDEBUG
  printf("\n bn\n");
  prtmatl(b, n, m);
#endif
  aplc_inittrs(res);
  res->type = APLC_REAL;
  res->rank = 2;
  aplc_vectalloc(&mptemp, 2, APLC_INT);
  res->shape = mptemp.ip;
  res->alloc_ind |= APLC_ALLOC_SHAPE_F;
  res->shape[0] = n;
  res->shape[1] = nrhs;
  res->size = n * nrhs;
  aplc_talloc(res);

  /* put transpose of (new) b in result */
  c = res->value;
  for (i = 0; i < n; i++) {
    for (j = 0; j < nrhs; j++) {
      *c.rp++ = (b.rp)[i + ldb * j];
    }
  }
  aplc_free(work);
  aplc_free(jpvt);/*xx*/
  aplc_free(b.rp);
  return;
}

#ifdef USE_SYSTEM_LAPACK
/* complex matrix inverse */
static void
matinv_z(struct trs_struct *res, double *ap, la_int m, la_int n)
{
  double rcond;
  double *work, *rwork;
  union mp_struct b;
  union mp_struct c;
  union mp_struct mptemp;
  int size, lsize;
  int i, j, k;
  la_int mn, nw;
  la_int *jpvt; 
  la_int nrhs, lda, ldb, rank, info;

  lda = m;
  ldb = (m < n) ? n : m;
  nrhs = m;
  size = m * n;
  rcond = DBL_EPSILON;

  /* setup workspace */
  mn = (m < n) ? m : n;
  nw = imaxl(mn + 3 * n, 2 * mn + nrhs);
  aplc_vectalloc(&mptemp, (int) 2*nw, APLC_REAL);
  work = mptemp.rp;
  aplc_vectalloc(&mptemp, (int) 2*n, APLC_REAL);
  rwork = mptemp.rp;

  jpvt = (la_int *) APLC_MALLOC( n*sizeof(la_int) );
  for (i = 0; i < n; ++i) {
    jpvt[i] = 0;
  }
  /* printf("\n mn %d, nw %d\n", mn, nw); */

  /* put identity in b  */
  lsize = ldb * nrhs;
  aplc_vectalloc(&b, 2*lsize, APLC_REAL);
  c = b;
  j = nrhs + 1;
  for (i = 0; i < lsize; i++) {
    if (0 == i % j)
      *c.rp++ = 1.0;
    else
      *c.rp++ = 0.0;
    *c.rp++ = 0.0;
  }

#if RMDEBUG
  /* sws  debugging printout */
  printf("\naplc_domino debugging\n");
  printf("\n a\n");
  prtmatz(ap,n,m);
  printf("\n b\n");
  prtmatz(b.rp,nrhs,ldb);
  printf("\n m %ld, n %ld, nrhs %ld, lda %ld, ldb %ld, rcond %g\n",
	 m,n,nrhs,lda,ldb, rcond);
#endif

  /* solve */
  zgelsx_(&m, &n, &nrhs, ap, &lda, b.rp, &ldb, jpvt, &rcond,
	  &rank, work, rwork, &info);
  /*zgelsy_(&m, &n, &nrhs, ap, &lda, b.rp, &ldb, jpvt, &rcond,
    &rank, work, rwork, &info);*/
  if (rank != n) {
    /* don't force exit for this error, just notify the user */
    fprintf(aplcerr, "\nzgelsx info = %ld, rank= %ld\n", (long) info, (long) rank);
    aplc_error("domain aplc_error [domino - matrix not of full right rank]");
  }
#if RMDEBUG
  printf("\n bn\n");
  prtmatz(b.rp, n, 2*m);
#endif
  aplc_inittrs(res);
  res->type = APLC_COMPLEX;
  res->rank = 2;
  aplc_vectalloc(&mptemp, 2, APLC_INT);
  res->shape = mptemp.ip;
  res->alloc_ind |= APLC_ALLOC_SHAPE_F;
  res->shape[0] = n;
  res->shape[1] = nrhs;
  res->size = n * nrhs;
  aplc_talloc(res);

  /* put transpose of (new) b in result */
  c = res->value;
  k = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < nrhs; j++) {
      c.rp[2*k] = (b.rp)[2*i + 2*ldb*j];
      c.rp[2*k+1] = (b.rp)[2*i + 2*ldb*j + 1];
      k++;
    }
  }
  aplc_free(work);
  aplc_free(rwork);
  aplc_free(jpvt);/*xx*/
  aplc_free(b.rp);
  return;
}
#endif /* #ifdef USE_SYSTEM_LAPACK */


/* matrix inverse */
extern void
aplc_domino(struct trs_struct * res, struct trs_struct *right)
{
  union mp_struct a;
  int type;
  int rrank;
  int *shape;
  int size;
  la_int m,n;

  type = right->type;
  rrank = right->rank;
  shape = right->shape;

  if (rrank == 2) {
    m = shape[0];
    n = shape[1];
  } else if (rrank == 1) {
    m = shape[0];
    n = 1;
  } else {
    m = 1;
    n = 1;
  }
  size = m * n;

  /* don't force exit for this error, just notify the user */
  if (m < n)
    aplc_error("length aplc_error [domino - non unique solution]");

  /* printf("\n mn %d, nw %d\n", mn, nw); */

  /* put stuff in transpose of a  */
  switch (type) {
  default:
    aplc_error("domain aplc_error [domino - int, real (complex) expected]");
    aplc_free(a.rp);
    aplc_zilde(res);
    return;
    break;

  case APLC_BOOL:
  case APLC_INT:
  case APLC_REAL:
    /* copy to real */
    aplc_vectalloc(&a, size, APLC_REAL);
    matrix_f_aplc(a.rp, right);
    matinv_real(res, a.rp, m,n);
    aplc_free(a.rp);
    break;

#ifdef USE_SYSTEM_LAPACK
  case APLC_COMPLEX:
    /* copy into transpose */
    aplc_vectalloc(&a, 2*size, APLC_REAL);
    matrix_z_f_aplc(a.rp, right);
    matinv_z(res, a.rp, m,n);
    aplc_free(a.rp);
    break;
#endif /* #ifdef USE_SYSTEM_LAPACK */
  }

  return;
}

/* solution to linear equations */
extern void
aplc_msolve(struct trs_struct *res,
	    struct trs_struct *left, struct trs_struct *right)
{
  union mp_struct a;
  union mp_struct b;
  union mp_struct c;
  union mp_struct mptemp;
  int ltype, lrank;
  int *lshape;
  int rtype, rrank;
  int *rshape;
  double *work, *rwork;
  int i, j, k;
  double rcond;
  int size, lsize;
  la_int mn, nw;
  la_int *jpvt; 
  la_int m, n, nrhs, lda, ldb, rank, info;
  int maxtype;

  rtype = right->type;
  rrank = right->rank;
  rshape = right->shape;
  ltype = left->type;
  lrank = left->rank;
  lshape = left->shape;

  if (rrank == 2) {
    m = rshape[0];
    n = rshape[1];
  } else if (rrank == 1) {
    m = rshape[0];
    n = 1;
  } else {
    m = 1;
    n = 1;
  }

  /* don't force exit for this error, just notify the user */
  if (m < n)
    aplc_error("length aplc_error [msolve - non unique solution]");

  if (lrank == 2) {
    ldb = lshape[0];
    nrhs = lshape[1];
  } else if (rrank == 1) {
    ldb = lshape[0];
    nrhs = 1;
  } else {
    ldb = 1;
    nrhs = 1;
  }
  if (m != ldb) {
    aplc_error("length aplc_error [msolve - dimensions don't match]");
    aplc_zilde(res);
    return;
  }
  lda = m;
  size = m * n;
  rcond = DBL_EPSILON;

  /* determine the result type */
  maxtype = imax( left->type, right->type );
  switch(maxtype) {
  default:
    aplc_error("domain aplc_error [msolve - int, real (complex) expected on right]");
    aplc_zilde(res);
    return;
    break;
  case APLC_BOOL:
  case APLC_INT:
  case APLC_REAL:
    /* setup workspace */
    mn = (m < n) ? m : n;
    nw = imaxl(mn + 3 * n, 2 * mn + nrhs);
    aplc_vectalloc(&mptemp, (int) nw, APLC_REAL);
    work = mptemp.rp;

    jpvt = (la_int *) APLC_MALLOC( n*sizeof(la_int) );
    for (i = 0; i < n; ++i) {
      jpvt[i] = 0;
    }
    /* printf("\n mn %d, nw %d\n", mn, nw); */
    /* put right stuff in transpose of a  */
    aplc_vectalloc(&a, size, APLC_REAL);
    matrix_f_aplc(a.rp, right);
    /* put transpose of left stuff in b  */
    lsize = ldb * nrhs;
    aplc_vectalloc(&b, lsize, APLC_REAL);
    matrix_f_aplc(b.rp, left);

#if RMDEBUG
    /* sws  debugging printout */
    printf("\naplc_msolve debugging\n");
    printf(" a\n");
    prtmatl(a,n,m);
    printf("\n b\n");
    prtmatl(b,nrhs,ldb);
    printf("\n m %ld, n %ld, nrhs %ld, lda %ld, ldb %ld, rcond %g\n",
           m,n,nrhs,lda,ldb, rcond);
#endif

    /* solve */
    dgelsx_(&m, &n, &nrhs, a.rp, &lda, b.rp, &ldb, jpvt, &rcond,
            &rank, work, &info);

    if (rank != n) {
      fprintf(aplcerr,"\ndgelsx info = %ld, rank= %ld\n", (long) info, (long) rank);
      aplc_error("domain aplc_error [msolve - matrix not of full right rank]");
    }
#if RMDEBUG
    printf("\n bn\n");
    prtmatl(b, nrhs, ldb);
#endif
    /* setup return value */
    aplc_inittrs(res);
    res->type = APLC_REAL;
    res->rank = 2;
    aplc_vectalloc(&mptemp, 2, APLC_INT);
    res->shape = mptemp.ip;
    res->alloc_ind |= APLC_ALLOC_SHAPE_F;
    res->shape[0] = n;
    res->shape[1] = nrhs;
    res->size = n * nrhs;
    aplc_talloc(res);

    /* put transpose of (new) b in result */
    c = res->value;
    for (i = 0; i < n; i++) {
      for (j = 0; j < nrhs; j++) {
        *c.rp++ = (b.rp)[i + ldb * j];
      }
    }
    break;

#ifdef USE_SYSTEM_LAPACK
  case APLC_COMPLEX:
    /* setup workspace */
    mn = (m < n) ? m : n;
    nw = imaxl(mn + 3 * n, 2 * mn + nrhs);
    aplc_vectalloc(&mptemp, (int) 2*nw, APLC_REAL);
    work = mptemp.rp;
    aplc_vectalloc(&mptemp, (int) 2*n, APLC_REAL);
    rwork = mptemp.rp;

    jpvt = (la_int *) APLC_MALLOC( n*sizeof(la_int) );
    for (i = 0; i < n; ++i) {
      jpvt[i] = 0;
    }
    /* printf("\n mn %d, nw %d\n", mn, nw); */
    /* put right stuff in transpose of a  */
    aplc_vectalloc(&a, 2*size, APLC_REAL);
    matrix_z_f_aplc(a.rp, right);
    /* put transpose of left stuff in b  */
    lsize = ldb * nrhs;
    aplc_vectalloc(&b, 2*lsize, APLC_REAL);
    matrix_z_f_aplc(b.rp, left);

#if RMDEBUG
    /* sws  debugging printout */
    printf("\naplc_msolve debugging\n");
    printf(" a\n");
    prtmatz(a.rp,n,m);
    printf("\n b\n");
    prtmatz(b.rp,nrhs,ldb);
    printf("\n m %ld, n %ld, nrhs %ld, lda %ld, ldb %ld, rcond %g\n",
           m,n,nrhs,lda,ldb, rcond);
#endif

    /* solve */
    zgelsx_(&m, &n, &nrhs, a.rp, &lda, b.rp, &ldb, jpvt, &rcond,
            &rank, work, rwork, &info);

    if (rank != n) {
      fprintf(aplcerr,"\ndgelsx info = %ld, rank= %ld\n", (long) info, (long) rank);
      aplc_error("domain aplc_error [msolve - matrix not of full right rank]");
    }
#if RMDEBUG
    printf("\n bn\n");
    prtmatz(b.rp, nrhs, ldb);
#endif
    /* setup return value */
    aplc_inittrs(res);
    res->type = APLC_COMPLEX;
    res->rank = 2;
    aplc_vectalloc(&mptemp, 2, APLC_INT);
    res->shape = mptemp.ip;
    res->alloc_ind |= APLC_ALLOC_SHAPE_F;
    res->shape[0] = n;
    res->shape[1] = nrhs;
    res->size = n * nrhs;
    aplc_talloc(res);

    /* put transpose of (new) b in result */
    c = res->value;
    k = 0;
    for (i = 0; i < n; i++) {
      for (j = 0; j < nrhs; j++) {
        c.rp[2*k] = (b.rp)[2*i + 2*ldb*j];
        c.rp[2*k+1] = (b.rp)[2*i + 2*ldb*j + 1];
        k++;
      }
    }
    break;
#endif /* #ifdef USE_SYSTEM_LAPACK */


  }

  aplc_free(work);
  aplc_free(jpvt);
  aplc_free(a.rp);
  aplc_free(b.rp);
  return;
}


/* general drop fn

   res .is left drop right

 */
extern void
aplc_drop(struct trs_struct * res,
	  struct trs_struct * left, struct trs_struct * right)
{
  int ltype, lrank;
  int *lshape;
  int rtype, rrank;
  int *rshape;
  union mp_struct mptemp;
  union mp_struct s, delta;
  int i,j;
  int degen_flag;
  int size;
  int alpha;
  int temp;
  int index;

  rtype = right->type;
  rrank = right->rank;
  rshape = right->shape;
  ltype = left->type;
  lrank = left->rank;
  lshape = left->shape;

  /* first some error checking */
  if ( (ltype != APLC_INT) && (ltype != APLC_BOOL) ) {
    fprintf(aplcerr,"left type = %d\n", ltype);
    aplc_error("[drop] type error (left not int)");
  }

#if 0
  fprintf(aplcerr,"[drop] lrank %d, rrank %d\n",lrank,rrank);  
  aplc_fprintf_ivec(aplcerr, "lshape", lshape, lrank);
  aplc_fprintf_ivec(aplcerr, "rshape", rshape, rrank);
#endif
  /* scalar should appear as a vector */
  if (lrank != 1)
    aplc_error("[drop] rank error (left not vector)");

  if (0 == rrank) {
    /* scalar right */
    res->type = rtype;
    res->rank = left->shape[0];
    /* allocate space for shape vector */
    aplc_vectalloc(&mptemp, res->rank, APLC_INT);
    res->shape = mptemp.ip;
    res->alloc_ind = APLC_ALLOC_SHAPE_F;
    /* shape components are 0 or 1 */
    degen_flag = 0;
    for (i=0; i<left->shape[0]; i++) {
      if (left->value.ip[i]) {
	res->shape[i] = 0;
	degen_flag = 1;
      } else {
	res->shape[i] = 1;
      }
    }
    if (degen_flag) {
      res->size = 0;
      res->value.ip = (void *)0;
    } else {
      res->size = aplc_vsize(res->rank, res->shape);
      aplc_talloc(res);
      aplc_cptrsval(res,0, right,0);/* set the value */
    }
    /*aplc_fprintf_ivec(aplcerr, "res->shape", res->shape, res->rank);*/
    return;
  }

  if (lshape[0] != rrank) {
    fprintf(aplcerr,"left shape[0] = %d, right rank = %d\n",
	    lshape[0],  rrank);
    aplc_error("[drop] shape error (left shape not right rank)");
  }

  /* setup result type, rank, shape */
  res->type = rtype;
  res->rank = rrank;
  /* allocate space for shape vector */
  aplc_vectalloc(&mptemp, rrank, APLC_INT);
  res->shape = mptemp.ip;
  res->alloc_ind = APLC_ALLOC_SHAPE_F;
  /* copy in left values to shape
     - also see if result is of 0 size
     - also compute offset vector, s
   */
  /* offset vector */
  aplc_vectalloc(&s, rrank, APLC_INT);
  degen_flag = 0;
  for (i=0; i<lshape[0]; i++) {
    res->shape[i] = right->shape[i] - aplc_iabs(left->value.ip[i]);
    if (0 >= res->shape[i]) {
      /* degenerate case, size is 0 */
      degen_flag = 1;
      res->shape[i] = 0;
    }
    if (left->value.ip[i] > 0)
      s.ip[i] = left->value.ip[i];
    else
      s.ip[i] = 0;
  }
  if (degen_flag) {
    /* we're done, don't need offset */
    aplc_free(s.ip);
    res->size = 0;
    return;
  }

  /* ok, to get here we must have something left    */
  alpha = aplc_td_accessor(res->rank, res->shape, 
			   rrank, right->shape,
			   &delta, s.ip);
#if 0
  /* print delta */
  printf("delta = [");
  for (i=0; i<rank; i++)
    printf("%d ",delta.ip[i]);
  printf("]\n");
#endif
  /* allocate space for result */
  res->size = aplc_vsize(res->rank, res->shape);
  size = aplc_talloc(res);
  index = alpha;
  for (i = 0; i < size; i++) {
    aplc_cptrsval(res,i, right,index);/* set the value */
    temp = res->shape[rrank-1];
    for (j = rrank - 1; j >= 0; j--) {
      index += delta.ip[j];
      if ( (j>0) && (0 == ((i + 1) % temp) ) )
	temp *= res->shape[j-1];
      else
	break;
    }
  }
  /* free the offset vector, delta */
  aplc_free(s.ip);
  aplc_free(delta.ip);
  return;
}

#define TDBG 0

/* general take fn

   res .is left take right

 */
extern void
aplc_take(struct trs_struct *res,
	  struct trs_struct *left, struct trs_struct *right)
{
  int ltype, lrank;
  int *lshape;
  int rtype, rrank;
  int *rshape;
  union mp_struct mptemp;
  union mp_struct s, delta;
  union mp_struct e, a, rse;
  int i,j;
  int degen_flag;
  int size;
  int alpha;
  int temp;
  int index;
  int lsize, rsize;
  union res_struct rfill; 
  int over_flag;
  int fill_flag;

  rtype = right->type;
  rrank = right->rank;
  rshape = right->shape;
  rsize = aplc_vsize(rrank, rshape);
  ltype = left->type;
  lrank = left->rank;
  lshape = left->shape;
  lsize = aplc_vsize(lrank, lshape);

#if TDBG
  printf("[take] left trs\n");
  aplc_print_trs(left);
  printf("[take] right trs\n");
  aplc_print_trs(right);
#endif
  /* first some error checking */
  /* left type must be an int */
  if ( (ltype != APLC_INT) && (ltype != APLC_BOOL) ) {
    fprintf(aplcerr,"left type = %d\n", ltype);
    aplc_error("[take] type error (left not int)");
  }
  /* left must be singleton or vector */ 
  if ( (lsize > 1) && (lrank != 1) )
    aplc_error("[take] rank error (left not vector)");

  /*  if (lshape[0] != rrank) {*/
  /* allow scalar right */ 
  if ( (0 < rrank) && (lshape[0] != rrank) ) {
    fprintf(aplcerr,"left shape[0] = %d, right rank = %d\n",
	    lshape[0],  rrank);
    aplc_error("[take] shape error (left shape not right rank)");
  }

  /* setup result type, rank, shape */
  res->type = rtype;
  if ( 0 < rrank )
    res->rank = rrank;
  else
    res->rank = lsize;
  /* allocate space for shape vector */
  aplc_vectalloc(&mptemp, res->rank, APLC_INT);
  res->shape = mptemp.ip;
  res->alloc_ind = APLC_ALLOC_SHAPE_F;
  /* copy in left values to shape */
  /* copy in left values to shape
     - also see if result is of 0 size
     - also compute offset vector, s
     - also see if we need overtake 
   */
  /* offset vector */
  aplc_vectalloc(&s, res->rank, APLC_INT);
  degen_flag = 0;
  over_flag = 0;
  /* note left shape[0] is right rank, should be the same as lsize    */ 
  for (i=0; i<lsize; i++) {
    res->shape[i] = aplc_iabs(left->value.ip[i]);
    if (0 >= res->shape[i]) {
      degen_flag = 1;
      res->shape[i] = 0;
    }
    if (left->value.ip[i] < 0) {
      if (rrank > 0)
	s.ip[i] = right->shape[i] + left->value.ip[i];/* note this subtracts */
      else
	s.ip[i] = 1 + left->value.ip[i];/* note this subtracts */
    } else
      s.ip[i] = 0;
    if ( rrank == 0 ) {
      if ( aplc_iabs(left->value.ip[i]) > 1) 
	over_flag = 1;
    } else if ( right->shape[i] < aplc_iabs(left->value.ip[i]) )
      over_flag = 1;
  }
  if (degen_flag) {
    /* we're done */
#if TDBG
    printf("[take] res d trs\n");
    aplc_print_trs(res);
#endif
    /* free the offset vector */
    aplc_free(s.ip);
    res->size = 0;
    return;
  }
  /* ok, to get here we must have something left    */
  alpha = aplc_td_accessor(res->rank, res->shape, 
			   rrank, right->shape,
			   &delta, s.ip);
  /* allocate space for result */
  res->size = aplc_vsize(res->rank, res->shape);
  size = aplc_talloc(res);

  if (0 == over_flag) {
    /* don't have to handle overtake */
    index = alpha;
    for (i = 0; i < size; i++) {
      /* set the value */
      aplc_cptrsval(res,i, right,index);
      temp = res->shape[res->rank-1];
      for (j = res->rank - 1; j >= 0; j--) {
	index += delta.ip[j];
	if ( (j>0) && (0 == ((i + 1) % temp) ) )
	  temp *= res->shape[j-1];
	else
	  break;
      }
    }
    /* free the offset vector, delta */
    aplc_free(s.ip);
    aplc_free(delta.ip);
#if TDBG
    printf("[take] res trs\n");
    aplc_print_trs(res);
#endif
    return;
  }

  /* ok, we need overtake

     s[i] is the offset for axis i, possibly negative

     use s as the index into right

   */

  /* build the exp vector for expanded right */
  aplc_vectalloc(&e, res->rank, APLC_INT);
  /* also build shape of expanded right */
  aplc_vectalloc(&rse, res->rank, APLC_INT);
  e.ip[res->rank-1] = 1;
  for (i = res->rank-1; i >0; i--) {
    /*e.ip[i-1] = right->shape[i]*e.ip[i];*/
    if (rrank > 0)
      rse.ip[i] = max( right->shape[i], aplc_iabs(left->value.ip[i]) );
    else
      rse.ip[i] = max( 1, aplc_iabs(left->value.ip[i]) );
    e.ip[i-1] = rse.ip[i]*e.ip[i];
  }
  if (rrank > 0)
    rse.ip[0] = max( right->shape[0], aplc_iabs(left->value.ip[0]) );
  else
    rse.ip[0] = max( 1, aplc_iabs(left->value.ip[0]) );

#if 0
  for (i=0; i< res->rank; i++) {
    printf("%d, %d\n", rse.ip[i], e.ip[i]);
  }
#endif

  /* storage for indicies into expanded right */
  aplc_vectalloc(&a, res->rank, APLC_INT);
  /* get the fill element */
  aplc_genfill(&rfill, rtype);
  index = alpha;
  for (i = 0; i < size; i++) {
    /* set the value
       - first determine indicies of expanded right
       - adjust with offset to right indicies
       - next fill if any are out of bounds
     */
    /* a[i] = (offset div e[i]) mod s[i] */
    fill_flag = 0;
    for (j = 0; j < res->rank; j++) {
      /* printf("i %d, j %d; %d, %d\n", i,j, e.ip[j], rse.ip[j]);*/
      a.ip[j] = (i/e.ip[j]) % rse.ip[j];
      /* adjust by offset */
      a.ip[j] = a.ip[j] + s.ip[j];
      if (rrank > 0) {
	if ( (a.ip[j] <0) || (a.ip[j] >= right->shape[j]) )
	  fill_flag = 1;
      } else {
	if ( (a.ip[j] <0) || (a.ip[j] >= 1) )
	  fill_flag = 1;
      }

    }
    /* now actually put the value or fill in */
    if ( fill_flag )
      aplc_cpres2trs(res,i, &rfill);
    else
      aplc_cptrsval(res,i, right,index);
    /* get next index */
    temp = res->shape[res->rank-1];
    for (j = res->rank - 1; j >= 0; j--) {
      index += delta.ip[j];
      if ( (j>0) && (0 == ((i + 1) % temp) ) )
	temp *= res->shape[j-1];
      else
	break;
    }
  }
  /* free the offset vector, delta */
  /* sometimes s.ip = a.ip; freed twice?? */
  aplc_free(e.ip);
  aplc_free(rse.ip);
  aplc_free(a.ip);
  aplc_free(s.ip);
  aplc_free(delta.ip);
#if TDBG
  printf("[take] res o trs\n");
  aplc_print_trs(res);
#endif
  return;
}

/* end of runmat.c */
